/* FILE: kl_pbc_util.h            -*-Mode: c++-*-
 * Function used by the periodic boundary conditions (PBC) demagnetization
 * modules (Klm_SimpleDemag_PBC, Klm_Demag_PBC).
 * N means demagnetization tensor (see paper of Newell, see demagcoef.h)
 * equations cited here are taken form this paper.
 
 In 1-2006 the Periodic Boundary Condition (PBC) functionality
 was introduced by Kristof Lebecki (nick KL(m)).
 With large help of Mike Donahue, of course.
 The (MD's) idea was, to expand the initialization part, where original
 interactions between cells were computed using formulae and
 functions defined in the paper of Newell.
 PBC concept is based on the idea that interaction between a
 cell and an (inifinite) series of cells leads to an (infinite) 
 *summation* of the coeffitients for the interation cellC-cellS,
 where cellC is our "central cell" and cellS belongs to the
 series of "repetitions" of an originall cellO:
   	cellS(n) = Zshift(cellO,n*zPeriod),  where
 	n = -Infinity, ..., -1, 0, 1, ..., +Infinity
 	zPeriod is a "repetition period"
 The summation \sum _{n=-Infinity} ^{+Infinity} N(cellC,cellS)
 (N is the demagnetization tensor) will be broken into three parts:
   1. "Image", shortcut *img*
      I will often denote it as cube-cube. 
      Central part (small values of |n|) will be calculated using 
      Newells cell-cell formulae. This part (area) is limited by the 
      cell-cell formulea, which lead to increasing computational errors 
      for large cell-cell distances. 
      This critical distance / radius is defined by parameters
      >>max_celldistance_img_*<<.
      I have evaluated this parameter for Wintel platform
      (8 byte double precision), and it apears to be in the
      order of 30.
      In the code I will express this critical distance also 
      with n-values (n is PBC-repetition counter).
   2. "Dipole", shortcut *dip*
      Near distance cellS's (medium values of |n|) are evaluated
      using the dipole estimation.
      This area is limited (large |n|) by the case where we are
      able to introduce an approximation of the *dipole*
      formula. Namely for large |n| the relative difference
      between the *approximation* and (somehow exact) *dipole*
      will be smaller than a value defined by a parameter
      >>err_ratio_cont_to_dip*error_img_dip_APPRIOPRIATE<<
   3. "Continous", shortcut *cont*
      Other names appearing int the code are: "Summation up to 
      the infinites" or "approximation".
      This approximation is choosen in such a nice way, that 
      we know the analytical formula for the summation with
      infinite upper (or lower) limit.
      For this *approximation*, we have choosen the integral of
      the *dipole* formulae.
      We will use formulae (functions) which give us ready
      results of the summation from an finite to infinite 
      limit. These results will have a form of a difference
      "finite boundary limit" - "infinite boundary limit"
 Each part (area) will use following functions:
   1. CalculateSDA**
   2. CalculateSDA**_dip
   3. CalculateSDA00_aprx_finZ (finite boundary limit)
      CalculateSDA00_aprx_infZ (infinite boundary limit)
  Concluding, we will sum several elements using way "1", "2"
  and add to it "infinite tails" using funtions **_aprx_**

 KL(m), 10-20-2005
 KL(m), 3-23-2006
   
Kristof M. Lebecki, 
Institute of Physics, Polish Academy of Sciences, dept. SL-3
Al. Lotnikow 32/46, 02-668 Warsaw, POLAND
phone: (48 22) 843 66 01 ext. 3193 / 2551
fax:   (48 22) 843 13 31
WWW:   www.ifpan.edu.pl/~lebecki
 */

#include "mesh.h"
#include "rectangularmesh.h"
#include "fft.h"		// just fot Oxs_Complex definition

/* End includes */

// Functions to shorten the code for large integer (positive) power.
inline REALWIDE pow2(REALWIDE x)
{ return x*x; }
inline REALWIDE pow3(REALWIDE x)
{ return x*x*x; }
inline REALWIDE pow4(REALWIDE x)
{ return x*x*x*x; }
inline REALWIDE pow5(REALWIDE x)
{ return x*x*x*x*x; }


// Functions for calculating the sphere / dipole N-values
// These functions can be succesfully used as a far-distance approximation
//   (replacement) for exact cube-cube Newell functions. 
// In the denominator, we write sqrt(X*X*X*X*X) instead of pow(X,2.5)
//	 for speed-up (30 times faster in IEEE environment, 5M/sec).
inline REALWIDE CalculateSDA00_dip(REALWIDE x,REALWIDE y,REALWIDE z)
{ // This is Nxx term (or N00) from eq. (15) without the mul-factor
  return (2.*x*x-y*y-z*z) / (3.*sqrt( pow5(x*x+y*y+z*z) ));
}
inline REALWIDE CalculateSDA11_dip(REALWIDE x, REALWIDE y, REALWIDE z)
{ return CalculateSDA00_dip(y,x,z); }
inline REALWIDE CalculateSDA22_dip(REALWIDE x, REALWIDE y, REALWIDE z)
{ return CalculateSDA00_dip(z,y,x); }
//
inline REALWIDE CalculateSDA01_dip(REALWIDE x,REALWIDE y,REALWIDE z)
{ // This is Nxy term (or N01) from eq. (15) without the mul-factor
  return x*y / sqrt( pow5(x*x+y*y+z*z) );
}
inline REALWIDE CalculateSDA02_dip(REALWIDE x, REALWIDE y, REALWIDE z)
{ return CalculateSDA01_dip(x,z,y); }
inline REALWIDE CalculateSDA12_dip(REALWIDE x, REALWIDE y, REALWIDE z)
{ return CalculateSDA01_dip(y,z,x); }


// Functions to calculate the *approximated* tails in infinities.
// Following functions reflect summation {from -\infty to -n_max} 
//  plus {from +n_max to +\infty}. For instance:
//   CalculateSDA00_aprx_tails ==   \sum _-\infty ^-n_max CalculateSDA00_dip
//                                 +\sum _+n_max ^-\infty CalculateSDA00_dip
REALWIDE CalculateSDA00_tails(REALWIDE x,REALWIDE y,REALWIDE z,
			REALWIDE zPeriod, UINT4m n_max);
REALWIDE CalculateSDA11_tails(REALWIDE x,REALWIDE y,REALWIDE z,
			REALWIDE zPeriod, UINT4m n_max);
REALWIDE CalculateSDA22_tails(REALWIDE x,REALWIDE y,REALWIDE z,
			REALWIDE zPeriod, UINT4m n_max);
REALWIDE CalculateSDA01_tails(REALWIDE x,REALWIDE y,REALWIDE z,
			REALWIDE zPeriod, UINT4m n_max);
REALWIDE CalculateSDA02_tails(REALWIDE x,REALWIDE y,REALWIDE z,
			REALWIDE zPeriod, UINT4m n_max);
REALWIDE CalculateSDA12_tails(REALWIDE x,REALWIDE y,REALWIDE z,
			REALWIDE zPeriod, UINT4m n_max);

// The above *approximated* functions return results with given known error.
// Following functions are a base which (to know the absolute error value) must be
//  multiplied by the required relative-error value (now I think it should be
//    err_ratio_cont_to_dip*error_img_dip_**diag)
REALWIDE CalculateSDA00_errorBase(REALWIDE x,REALWIDE y,REALWIDE z,
			REALWIDE zPeriod, UINT4m n_max);
REALWIDE CalculateSDA11_errorBase(REALWIDE x,REALWIDE y,REALWIDE z,
			REALWIDE zPeriod, UINT4m n_max);
REALWIDE CalculateSDA22_errorBase(REALWIDE x,REALWIDE y,REALWIDE z,
			REALWIDE zPeriod, UINT4m n_max);
REALWIDE CalculateSDA01_errorBase(REALWIDE x,REALWIDE y,REALWIDE z,
			REALWIDE zPeriod, UINT4m n_max);
REALWIDE CalculateSDA02_errorBase(REALWIDE x,REALWIDE y,REALWIDE z,
			REALWIDE zPeriod, UINT4m n_max);
REALWIDE CalculateSDA12_errorBase(REALWIDE x,REALWIDE y,REALWIDE z,
			REALWIDE zPeriod, UINT4m n_max);

// Determinant of a symmetric tensor 
REAL8m determinant( \
  	REALWIDE nXX, REALWIDE nYY, REALWIDE nZZ, 
  	REALWIDE nXY, REALWIDE nXZ, REALWIDE nYZ);

// Assuming a Nij element of the tensor N is known with an absolute error dNij
//  how would it influence the vector H=N*m (for any vector m)?
// Answer: *maximum* relative error of the vector H: dH/H is given by *following* functions
//  multiplied by dNij.
// Example for N01: dH/H = dN01*rel_err_factor_01
REAL8m rel_err_factor_01( \
  	REALWIDE nXX, REALWIDE nYY, REALWIDE nZZ, 
  	REALWIDE nXY, REALWIDE nXZ, REALWIDE nYZ, REAL8m det);
inline REAL8m rel_err_factor_02( \
  	REALWIDE nXX, REALWIDE nYY, REALWIDE nZZ, 
  	REALWIDE nXY, REALWIDE nXZ, REALWIDE nYZ, REAL8m det)
// we exchange tensor columns: 2 & 3 and rows: 2 & 3  	
{ return rel_err_factor_01(nXX, nZZ, nYY, nXZ, nXY, nYZ, det); }
inline REAL8m rel_err_factor_12( \
  	REALWIDE nXX, REALWIDE nYY, REALWIDE nZZ, 
  	REALWIDE nXY, REALWIDE nXZ, REALWIDE nYZ, REAL8m det)
// we exchange tensor columns: 1 & 2 and rows: 1 & 2  	
{ return rel_err_factor_01(nYY, nXX, nZZ, nXY, nYZ, nXZ, det); }

  	      
// used ex. in the CalculateSDA_PBC_really function
// tensor element indicator
enum  T_tens_el { n00, n11, n22, n01, n02, n12 }; // !! update tens_el_MAX !!
const UINT2m tens_el_MAX=6; // number of elements !! keep this value up-to-date

// tensor_file relation
enum FileAccessType { WRITE, READ, NOTHING };

// execution order in the *calling* *structure*
// _SimpleDemag means that first A00 are computed (for all cells), 
//	then A11, A22, A01, A02, and A12.
// _Demag means that first A00, A01, and A02 are computed (for all cells
//	*in parallel*), then A11, A12, A22.
// This order affects:
//	- progress indication (done during computations)
//	- tensor file storage (done during computations and in I/O methods)
enum  T_exec_order { _SimpleDemag, _Demag }; // !! update exec_order_MAX !!
const UINT2m exec_order_MAX=2; // number of elements 
                               // !! keep this value up-to-date

class DemagPBC {
  private:
    // Constructor parameters with event. small deviations
    // meshes, both are needed
    // declared const as such are here brought
    const Oxs_Mesh* genmesh;
    const Oxs_RectangularMesh* mesh;
    // z-period
    mutable REALWIDE zPeriod;   // in relative "dz" units
    // ...
    const REAL4m max_celldistance_img_diag;
    const REAL4m max_celldistance_img_OFFdiag;
    // ...
    const REAL4m error_img_dip_diag;
    const REAL4m error_img_dip_OFFdiag;
    // ...
    const REAL4m err_ratio_cont_to_dip;  
    // ...
    const UINT4m max_no_of_sum_elements;
    // ...
    const INT4m max_no_of_sum_elements_debug;
    // tensor_file - is not needed to be seen globally, we will evaluate it in
    //               the constructor. See below for associated variables.
    const   String tensor_file_suffix;
    // progress script is not needed (to be seen globally). See below, as well...
    // cyclic convolutions are not as well
    const   T_exec_order exec_order;
    
    // mesh parameters
    mutable UINT4m xdim, ydim, zdim; // natural size
    mutable UINT4m xydim;
    mutable UINT4m totalsize;
    // relative cell edge sizes
    mutable REALWIDE dx, dy, dz;
    
    // tensor_file-specific
    mutable FileAccessType tensor_rw; // read or write or else
    mutable String tensor_file_prefix;

    // progress-specific
    const String progress_script;
    mutable Tcl_Interp* interp;       // for Tcl_Eval
    mutable BOOL show_progress;       // Yes or No 
    mutable UINT4m progress_total;    // denominator, "how much to do"
    mutable UINT4m progress_cnt;      // numerator, "actual done work"

   // maximal numbers of repetitions after which the difference apprx-dip should
   //  be below the "err_ratio_cont_to_dip*error_img_dip_APPRIOPRIATE" level
    mutable UINT4m n_max[tens_el_MAX];
    // error introduced by the *approximation*
    mutable REAL8m max_apprx_err[tens_el_MAX]; // Relative error
    mutable REAL8m sum_apprx_err[tens_el_MAX]; // Sum, used later for averaging
    mutable UINT4m num_apprx_err[tens_el_MAX]; // Used for enumarating (to average)
    // Relative errors introduced by the *approximation*, and also by *dip* and *img*
    mutable REAL8m max_tot_err[tens_el_MAX];   // Maximal value
    mutable REAL8m sum_tot_err[tens_el_MAX];   // Sum, used later for averaging
    mutable UINT4m num_tot_err[tens_el_MAX];   // Used for enumarating (to average)
    
    // Buffers for error statistics calculation
    // Used in case, when (tensor_rw==WRITE || tensor_rw==NOTHING)
    // All are modified inside of CalculateSDA_PBC_really
    // "infinity tails" errors (absolute, multiplied already by err_ratio_cont_to_dip*error_img_dip_)
    mutable Oxs_MeshValue<REAL4m> A00_err_inf_tail;
    mutable Oxs_MeshValue<REAL4m> A11_err_inf_tail;
    mutable Oxs_MeshValue<REAL4m> A22_err_inf_tail;
    mutable Oxs_MeshValue<REAL4m> A01_err_inf_tail;
    mutable Oxs_MeshValue<REAL4m> A02_err_inf_tail;
    mutable Oxs_MeshValue<REAL4m> A12_err_inf_tail;
    // "sum_coef" values
    mutable Oxs_MeshValue<REAL4m> A00_sum_coef;
    mutable Oxs_MeshValue<REAL4m> A11_sum_coef;
    mutable Oxs_MeshValue<REAL4m> A22_sum_coef;
    mutable Oxs_MeshValue<REAL4m> A01_sum_coef;
    mutable Oxs_MeshValue<REAL4m> A02_sum_coef;
    mutable Oxs_MeshValue<REAL4m> A12_sum_coef;
    // Tensor file storage buffer
    // Used in case, when (tensor_rw==WRITE || tensor_rw==READ)
    mutable Oxs_MeshValue<ThreeVector> tensor_buffer;  
    
    // methods ********************************
    
    void ReleaseMemory() const; // Conceptually const.  
    
    // Function to calculate the tensor element (coef in originall SimpleDemag)
    // Inputs are various variables like position (xyz), size (dx...), etc.
    // Some parameters are taken from the class body.
    // Inf_tails and sum_coeff are written on request to appr tables 
    // (with index store_inf_sum_idx)
    // conceptually const
    REALWIDE CalculateSDA_PBC_really( \
        const UINT4m i, const UINT4m j, const UINT4m k, \
        const T_tens_el teXX) const;
        
    // all tensor_buffer accesses should use following methods
    // (as Mike plans to change some OOMMF code somaday)
    REAL8m GetTensorElementFromBuffer(
        const UINT4m el_idx, const T_tens_el teXX) const;
    // conceptually const
    void PutTensorElementToBuffer(
        const REAL8m elem, const UINT4m el_idx, const T_tens_el teXX) const;
        
    // Upload/download routines
    // conceptually conts, changes the tensor_buffer
    BOOL load_tensor_file( const String file_name, 
                           const String title) const;
    // really const
    void save_tensor_file( const String file_name, 
                           const String title,
                           const char* desc) const;
                                                      
  public:
    // constructor
    // Mainly imports the parameters and does small evaluation of them,
    // initializes the variables
    DemagPBC( const Oxs_Mesh* genmesh,
              const UINT4m zPeriod_param,
              const REAL4m max_celldistance_img_diag,
              const REAL4m max_celldistance_img_OFFdiag,
              const REAL4m error_img_dip_diag,
              const REAL4m error_img_dip_OFFdiag,
              const REAL4m err_ratio_cont_to_dip,  
              const UINT4m max_no_of_sum_elements,
              const INT4m max_no_of_sum_elements_debug,
              const String tensor_file_name,
              const String tensor_file_suffix,
              const String progress_script,
              const BOOL cyclic_convolution,
              const T_exec_order exec_order
            );
    
    // Written separately as errors thorw here are better evaluated
    //  as those thrown from inside the constructor.
    // All memory allocation code is here.
    // conceptually const
    void Init() const;
    
    // computation of the tensor for the given cell
    // conceptually const                       
    REALWIDE CalculateSDA(const T_tens_el teXX,
                  const UINT4m i, const UINT4m j, const UINT4m k) const;
                  
    // Tensor_file upload/download.
    // Should be run in the middle of computations: after 00, 11, 22
    //  and before 01, 02, 12.
    // conceptually conts, changes the tensor_buffer
    void middle_load_or_save_tensor_file() const;
          
    // final step
    // The Aij tables are only for error statistics evaluation (and are 
    // important for off-diagonal element errors).
    // If you do not know them (like in Oxs_Demag case), place NULL instead
    // conceptually const
    void Finalize(
        const Oxs_Complex *A00, const Oxs_Complex *A11, const Oxs_Complex *A22, 
        const Oxs_Complex *A01, const Oxs_Complex *A02, const Oxs_Complex *A12
      ) const;
        
    // destructor
    ~DemagPBC() { ReleaseMemory(); }
    
    
};
